"""
@Author: rishi
"""
# Import required modules
import random as rd


# Build List class
class List:
    # Creates random list
    @staticmethod
    def create_random_list(length=10):
        """
        :param length: length of the list
        :return: list of random numbers
        """
        return [rd.randint(0, length) for _ in range(length)]

    # Create random 2d list
    @staticmethod
    def create_random_2d_list(row_length=4, col_length=4):
        """
        :param row_length: row length of the list
        :param col_length: col length of the list
        :return: 2d list of random numbers
        """
        return [[rd.randint(0, 10) for _ in range(row_length)] for _ in range(col_length)]



"""
@Author: rishi
"""


class TreeNode:
    def __init__(self, val):
        self.__val = val
        self.__left = None
        self.__right = None

    def get_val(self):
        return self.__val

    def get_left(self):
        return self.__left

    def get_right(self):
        return self.__right

    def set_val(self, val):
        self.__val = val

    def set_left(self, left):
        self.__left = left

    def set_right(self, right):
        self.__right = right

    def insert_node(self, val):
        # Check for root node is none
        if self is None:
            self = TreeNode(val)
            return
        else:
            if val == self.__val:
                return
            # Check val is less than root.val
            elif val < self.__val:
                if self.__left:
                    self.__left.insert_node(val)
                else:
                    self.__left = TreeNode(val)
                    return
            else:
                if self.__right:
                    self.__right.insert_node(val)
                else:
                    self.__right = TreeNode(val)
                    return



"""
@Author: rishi
"""
# Import required modules
import random

from .node import Node


# Build linked list class
class LinkedList:
    # Constructor
    def __init__(self):
        self.__head = None

    # Getter for head node
    def get_head(self):
        return self.__head

    # Check whether the linked list is empty
    def is_empty(self):
        return self.get_head() is None

    # Creates and inserts node into the linked list
    def insert_node(self, val):
        """
        :param val: value of the Node to be created
        """

        # if head Node is none then create the head Node with the value
        if self.is_empty():
            self.__head = Node(val)

        else:
            # create a copy of head Node
            temp = self.get_head()

            # traverse the available linked list
            while temp.get_next() is not None:
                temp = temp.get_next()

            # Add Node at the end of linked list
            temp.set_next(val)

    # Prints the linked list
    @staticmethod
    def pretty_print(headnode):
        """
        :param headnode: Head Node of linked list class instance
        """

        __str = ""
        head = headnode

        while head is not None:
            __str += str(head.get_val()) + "->"
            head = head.get_next()
        __str += "None"
        print(__str)

    # Creates a linked list with the values in array
    def create_linked_list(self, array):
        """
        :param array: iterable that specifies the values of linked list to be created
        :return: On success returns head node of linked list
        """

        # Traverse the array
        for i in array:
            self.insert_node(i)

        # Return success
        head = self.get_head()
        return head

    # Creates random linked list
    @staticmethod
    def create_random_linked_list(length=10):
        """
        :param length: length and range of the linked list
        :return: head node of the linked list

        """
        obj = LinkedList()
        for i in range(length):
            obj.insert_node(random.randint(1, length))
        return obj.get_head()



"""
@Author: rishi
"""

__version__ = '1.0.0'

# Import required modules
from QuickDS.implementations.binarysearchtree import BinarySearchTree as bst
from QuickDS.implementations.inputformatter import InputFormatter
from QuickDS.implementations.linkedlist import LinkedList
from QuickDS.implementations.list import List


# Main functions for linked list class


# Create linked list from array
def create_linked_list(array):
    """
    :param array: list of integers
    :return: head node of linked list
    """
    return LinkedList().create_linked_list(array)


# Create random linked list of given size or of size 10
def create_random_linked_list(length=10):
    """
    :param length: length of the linked list and range of the items in linked list
    :return: head of linked list
    """
    return LinkedList().create_random_linked_list(length)


# Print linked list
def pretty_print_llist(head):
    """
    :param head: head node of the linked list to be printed
    """
    LinkedList().pretty_print(head)


# Main functions for list class

# Creates random list
def create_random_list(length=10):
    """
    :param length: length of the list and range of the items in list
    :return: random list
    """
    return List().create_random_list(length)


# Creates 2d random list
def create_random_2d_list(row_length=4, col_length=4):
    """
    :param row_length: length of the rows and range of the items in row of list
    :param col_length: length of the rows and range of the items in row of list
    :return: random 2d list
    """
    return List().create_random_2d_list(row_length, col_length)


# Main functions for Input Formatter class


# Formats 1d char array
def format_1d_char_list(inp):
    """
    :param inp: string with 1d char array
    :return: 1d list
    """
    return InputFormatter().format_1d_char_list(inp)


# Formats 1d int array
def format_1d_int_list(inp):
    """
    :param inp: string with 1d int array
    :return: 1d list
    """
    return InputFormatter().format_1d_int_list(inp)


# Formats 2d int array
def format_2d_int_list(inp, row_length, col_length):
    """
    :param inp: string with 2d int array
    :param row_length: length of the 2d array rows
    :param col_length: length of the 2d array columns
    :return: 2d list
    """
    return InputFormatter().format_2d_int_list(inp, row_length, col_length)


# Formats 2d char array
def format_2d_char_list(inp, row_length, col_length):
    """
    :param inp: string with 2d int array
    :param row_length: length of the 2d array rows
    :param col_length: length of the 2d array columns
    :return:
    """
    return InputFormatter().format_2d_char_list(inp, row_length, col_length)


# Formats linked list
def format_linked_list(inp):
    """
    :param inp: string with int array
    :return: head node of linked list
    """
    return InputFormatter().format_linked_list(inp)


# Formats BST
def format_bst(inp):
    """
    :param inp: input string
    :return: root node of the tree
    """
    return InputFormatter().format_bst(inp)


# Main functions for Binary Search Tree class


# Creates random list
def create_random_binary_search_tree(no_of_nodes=10):
    """
    :param no_of_nodes: length of the list and range of the items in list
    :return: random list
    """
    return bst().create_random_binary_search_tree(no_of_nodes)


# Creates bst from sorted list
def convert_sorted_list_to_bst(array):
    """
    :param array: sorted array
    :return: root node of the bst created
    """
    return bst().convert_sorted_list_to_bst(array, 0, len(array) - 1)


# Converts bst to encoded string
def serialize(root):
    """
    :param root: root node of the tree
    :return: encoded string
    """
    return bst.serialize(root)


# Converts encoded string to BST
def deserialize(inp):
    """
    :param inp: encoded string
    :return: root node of the BST
    """
    return bst.deserialize(inp)


# Prints BST in viewable format
def pretty_print_bst(root):
    """
    :param root: root node of the BST
    """
    bst().pretty_print("", root, False)


# Print inorder of the BST
def print_bst(root):
    """
    :param root: root node of the tree
    """
    bst().inorder(root)



"""
@Author: rishi
"""

# Import required modules
import implementations.linkedlist as llist
import implementations.binarysearchtree as bst


# Class to build InputFormatter
class InputFormatter:
    # Returns matrix from string
    @staticmethod
    def __matrix_from_string(inp, row_length, col_length):
        """
        :param inp: string
        :param row_length: row length of matrix
        :param col_length: column length of matrix
        :return: matrix
        """

        # Split string into char array
        inp = inp.split(",")

        # Declare 2d matrix
        matrix = [[0] * col_length for _ in range(row_length)]

        ind = 0  # Index variable to traverse inp array

        for i in range(row_length):
            for j in range(col_length):
                matrix[i][j] = int(inp[ind])
                ind += 1
        return matrix

    # Formats 1d int array
    @staticmethod
    def format_1d_int_list(inp):
        """
        :param inp: string with 1d int array
        :return: 1d list
        """
        inp = inp.replace("{", "").replace("}", "")
        inp = inp.replace("[", "").replace("]", "")
        inp = inp.split(",")
        # convert char list to int list
        return list(map(lambda a: int(a), inp))

    # Formats 1d char array
    @staticmethod
    def format_1d_char_list(inp):
        """
        :param inp: string with 1d char array
        :return: 1d list
        """
        inp = inp.replace("{", "").replace("}", "").replace("'", "")
        inp = inp.replace("[", "").replace("]", "")
        inp = inp.split(",")
        # convert char list to int list
        return list(map(lambda a: int(a), inp))

    # Formats 2d int array
    @staticmethod
    def format_2d_int_list(inp, row_length, col_length):
        """
        :param inp: string with 2d int array
        :param row_length: length of the 2d array rows
        :param col_length: length of the 2d array columns
        :return: 2d list
        """
        inp = inp.replace("{", "").replace("}", "")
        inp = inp.replace("[", "").replace("]", "")
        return InputFormatter.__matrix_from_string(inp, row_length, col_length)

    # Formats 2d char array
    @staticmethod
    def format_2d_char_list(inp, row_length, col_length):
        """
        :param inp: string with 2d int array
        :param row_length: length of the 2d array rows
        :param col_length: length of the 2d array columns
        :return: 2d list
        """
        inp = inp.replace("{", "").replace("}", "").replace("'", "")
        inp = inp.replace("[", "").replace("]", "")
        return InputFormatter.__matrix_from_string(inp, row_length, col_length)

    # Formats linked list
    @staticmethod
    def format_linked_list(inp):
        """
        :param inp: string with int array
        :return: head node of linked list
        """
        inp = InputFormatter.format_1d_int_list(inp)
        return llist.LinkedList().create_linked_list(inp)

    # Formats BST
    @staticmethod
    def format_bst(inp):
        """
        :param inp: input string
        :return: root node of the tree
        """
        inp = inp.replace("[", "").replace("]", "")
        inp = inp.replace("{", "").replace("}", "")
        if len(inp) == 0:
            return None
        return bst.BinarySearchTree().deserialize(inp)



"""
@Author: rishi
"""

# Import required modules
import collections

from implementations.list import List
from implementations.treenode import TreeNode


# Build BST class
class BinarySearchTree:
    # Constructor
    def __init__(self):
        self.__root = None

    # Getters and Setters for root
    def get_root(self):
        return self.__root

    def set_root(self, root):
        self.__root = root

    # Check for root node is none
    def is_root(self):
        return self.__root is None

    # Insert a Node into BST
    def __insert(self, val):
        """
        :param val: value of the nodw to be inserted
        """
        if self.__root is None:
            self.__root = TreeNode(val)
        else:
            self.__root.insert_node(val)

    # Creates Random BST
    def create_random_binary_search_tree(self, length=10):
        """
        :param length: length of the array
        :return: root node of the random bst
        """
        # Create random array of given length
        arr = sorted(List.create_random_list(length))
        # Convert array to bst
        return self.convert_sorted_list_to_bst(arr, 0, length - 1)

    # Create sorted list to bst
    def convert_sorted_list_to_bst(self, arr, lower_bound, upper_bound):
        """
        :param arr: sorted list
        :param lower_bound: starting index value
        :param upper_bound: Ending index value
        :return: root node of the bst created
        """
        # Check whether lower_bound is less than upper_bound
        if upper_bound < lower_bound:
            return None
        else:
            # find middle index
            middle = lower_bound + (upper_bound - lower_bound) // 2

            # Create Node with middle index
            root = TreeNode(arr[middle])

            # Recurse to create nodes in left and right sub tree
            root.set_left(self.convert_sorted_list_to_bst(arr, lower_bound, middle - 1))
            root.set_right(self.convert_sorted_list_to_bst(arr, middle + 1, upper_bound))

            # Return root node of the tree
            return root

    # Convert BST to encoded string
    @staticmethod
    def serialize(root):
        """
        :param root: root node of the tree
        :return: encoded string
        """

        # If root is none return empty string
        if not root:
            return ''

        # Create a queue[FIFO] to store values
        q = collections.deque([root])
        res = []

        while q:
            node = q.popleft()
            # Add node value to array
            if node:
                res.append(str(node.get_val()))
                q.append(node.get_left())
                q.append(node.get_right())
            if not node:
                res.append('null')
        # Convert array to string and return
        return ','.join(res)

    # Convert encoded string to BST
    @staticmethod
    def deserialize(data):
        """
        :param data: serialized string
        :return: root node to bst
        """

        # If data is empty return none
        if not data:
            return None

        index = 1
        tem = data.split(',')
        root = TreeNode(tem[0])

        # Create queue to store values
        q = collections.deque([root])

        while q:
            node = q.popleft()

            # Add values to BST
            if tem[index] != 'null':
                node.set_left(TreeNode(tem[index]))
                q.append(node.get_left())
            index += 1
            if tem[index] != 'null':
                node.set_right(TreeNode(tem[index]))
                q.append(node.get_right())
            index += 1
        return root

    # Prints the bst in inorder(left_subtree->root->right_subtree)
    def inorder(self, root):
        """
        :param root: root node of the bst to be printed
        """

        # If root is not None
        if root:
            # Recurse left subtree
            self.inorder(root.get_left())
            # Print root value
            print(root.get_val(), end=" ")
            # Recurse right subtree
            self.inorder(root.get_right())

    # Prints the bst to view the tree
    def pretty_print(self, prefix, root, is_left):
        if root is not None:
            print(prefix + ("|-- " if is_left else "\\-- ") + str(root.get_val()))
            self.pretty_print(prefix + ("|   " if is_left else "    "), root.get_left(), True)
            self.pretty_print(prefix + ("|   " if is_left else "    "), root.get_right(), False)



'''
@Author: rishi
'''
import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="QuickDS",
    version="1.0.0",
    author="Rishi Raj",
    author_email="rishi772001@gmail.com",
    description="QuickDS is a library that contains a curated list of Input Formatting and Input Generating Functions that can be used to generate some commonly used Abstract data types without having to write any code from scratch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/varnaa/QuickDS/tree/master/python",
    license = "MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    include_package_data = True,
    python_requires='>=3.6',
    project_urls={  # Optional
            'Source': 'https://github.com/varnaa/QuickDS',
        }
)



"""
@Author: rishi
"""


# Class to build Linked List nodes
class Node:
    # Constructor
    def __init__(self, val):
        self.__val = val
        self.__next = None

    # Getters ans Setters for Node class members
    def get_next(self):
        return self.__next

    def get_val(self):
        return self.__val

    def set_next(self, val):
        self.__next = Node(val)



"""
@Author: rishi
"""

from QuickDS import *

# Test Linked list class
print("------LinkedList------")
array = [1, 2, 3, 4, 5]
pretty_print_llist(create_linked_list(array))
pretty_print_llist(create_random_linked_list(20))

# Test List class
print("------List------")
print(create_random_list(20))
print(create_random_2d_list(5, 5))

# Test inputformatter class
print("------InputFormatter------")
print(format_1d_int_list("{1, 2, 3}"))
print(format_1d_char_list("{'1', '2', '3'}"))
print(format_2d_int_list("{{1, 2, 3}, {4, 5, 6}}", 2, 3))
print(format_2d_char_list("{{'1', '2', '3'}, {'4', '5', '6'}}", 2, 3))
pretty_print_llist(format_linked_list("{1, 2, 3}"))
pretty_print_bst(format_bst("[3,1,4,null,2,null,5,null,null,null,null]"))

# Test BinarySearchTree class
print("------BST------")
r1 = convert_sorted_list_to_bst([1, 2, 3, 4, 5])
st = serialize(r1)
print(st)
r2 = deserialize(st)
r3 = create_random_binary_search_tree()
pretty_print_bst(r1)
pretty_print_bst(r3)
print_bst(r2)
